@@ -1,8 +1,10 @@
 #![doc(hidden)]
-use crate::classes::*;
+#![allow(unused_variables)]
+use crate::device_class::*;
 use crate::extra::*;
 use crate::private::Properties;
 use crate::Device;
+use crate::Fetch;
 use std::path::PathBuf;
 
 /// This is where PCI devices are located.
@@ -32,22 +34,24 @@ pub struct LinuxPCIDevice {
 
 impl Device for LinuxPCIDevice {
     fn new(path: &str) -> Self {
-        let mut path_vec = [path].to_vec();
         let mut device: LinuxPCIDevice = Default::default();
+        let mut path_vec = [path].to_vec();
 
         // One of the following two conditions will try to autocomplete the path of the
         // PCI device if the one provided doesn't point to a real path in the filesystem.
-        //   e.g.  0000:00:00.0 ->  /sys/bus/pci/devices/0000:00:00.0
-        //         00:00.0      ->  /sys/bus/pci/devices/0000:00:00.0
         if !PathBuf::from(path_vec.concat()).is_dir() {
+            // e.g. 0000:00:00.0 ->  /sys/bus/pci/devices/0000:00:00.0
             path_vec.insert(0, PATH_TO_PCI_DEVICES);
-            device.path = PathBuf::from(path_vec.concat());
+            device.set_path(PathBuf::from(path_vec.concat()));
             if !PathBuf::from(path_vec.concat()).is_dir() {
+                // e.g. 00:00.0      ->  /sys/bus/pci/devices/0000:00:00.0
                 let mut id = path.to_owned();
                 id.insert_str(0, "0000:");
                 std::mem::swap(&mut path_vec[1], &mut id.as_str());
-                device.path = PathBuf::from(path_vec.concat());
+                device.set_path(PathBuf::from(path_vec.concat()));
             }
+        } else {
+            device.set_path(PathBuf::from(path_vec.concat()));
         }
 
         device.set_address();
@@ -77,20 +81,20 @@ impl Device for LinuxPCIDevice {
         self.address.to_owned()
     }
 
-    fn class_id(&self) -> String {
-        hex::encode(self.class_id.to_owned())
+    fn class_id(&self) -> Vec<u8> {
+        self.class_id.to_owned()
     }
 
-    fn vendor_id(&self) -> String {
-        hex::encode(self.vendor_id.to_owned())
+    fn vendor_id(&self) -> Vec<u8> {
+        self.vendor_id.to_owned()
     }
 
-    fn device_id(&self) -> String {
-        hex::encode(self.device_id.to_owned())
+    fn device_id(&self) -> Vec<u8> {
+        self.device_id.to_owned()
     }
 
     fn numa_node(&self) -> isize {
-        self.numa_node.to_owned()
+        self.numa_node
     }
 
     fn class_name(&self) -> String {
@@ -117,24 +121,53 @@ impl Device for LinuxPCIDevice {
         self.d3cold_allowed
     }
 
-    fn revision(&self) -> String {
-        hex::encode(self.revision.to_owned())
+    fn revision(&self) -> Vec<u8> {
+        self.revision.to_owned()
     }
 
     fn subsystem_name(&self) -> String {
         self.subsystem_name.to_owned()
     }
 
-    fn subsystem_vendor_id(&self) -> String {
-        hex::encode(self.subsystem_vendor_id.to_owned())
+    fn subsystem_vendor_id(&self) -> Vec<u8> {
+        self.subsystem_vendor_id.to_owned()
     }
 
-    fn subsystem_device_id(&self) -> String {
-        hex::encode(self.subsystem_device_id.to_owned())
+    fn subsystem_device_id(&self) -> Vec<u8> {
+        self.subsystem_device_id.to_owned()
     }
 }
 
 impl Properties for LinuxPCIDevice {
+    fn reserved_new(path: &str) -> Self {
+        let mut device: LinuxPCIDevice = Default::default();
+        let mut path_vec = [path].to_vec();
+
+        // One of the following two conditions will try to autocomplete the path of the
+        // PCI device if the one provided doesn't point to a real path in the filesystem.
+        if !PathBuf::from(path_vec.concat()).is_dir() {
+            // e.g. 0000:00:00.0 ->  /sys/bus/pci/devices/0000:00:00.0
+            path_vec.insert(0, PATH_TO_PCI_DEVICES);
+            device.set_path(PathBuf::from(path_vec.concat()));
+            if !PathBuf::from(path_vec.concat()).is_dir() {
+                // e.g. 00:00.0  ->  /sys/bus/pci/devices/0000:00:00.0
+                let mut id = path.to_owned();
+                id.insert_str(0, "0000:");
+                std::mem::swap(&mut path_vec[1], &mut id.as_str());
+                device.set_path(PathBuf::from(path_vec.concat()));
+            }
+        } else {
+            device.set_path(PathBuf::from(path_vec.concat()));
+        }
+
+        // reserved_new tries to fetch the least amount of data at first.
+        // All the other fields can be populated later on.
+        device.set_class_id();
+        device.set_class_name();
+
+        device
+    }
+
     fn set_path(&mut self, p: PathBuf) {
         self.path = p;
     }
@@ -213,58 +246,59 @@ impl Properties for LinuxPCIDevice {
     }
 
     fn set_class_name(&mut self) {
+        if self.class_id.is_empty() {
+            return;
+        }
+
         // Associate class_id with class_name
-        if !&self.class_id.is_empty() {
-            self.class_name = match &self.class_id[0] {
-                1 => DeviceClass::MassStorageController.to_string(),
-                2 => DeviceClass::NetworkController.to_string(),
-                3 => DeviceClass::DisplayController.to_string(),
-                4 => DeviceClass::MultimediaController.to_string(),
-                5 => DeviceClass::MemoryController.to_string(),
-                6 => DeviceClass::Bridge.to_string(),
-                7 => DeviceClass::CommunicationController.to_string(),
-                8 => DeviceClass::GenericSystemPeripheral.to_string(),
-                9 => DeviceClass::InputDeviceController.to_string(),
-                10 => DeviceClass::DockingStation.to_string(),
-                11 => DeviceClass::Processor.to_string(),
-                12 => DeviceClass::SerialBusController.to_string(),
-                13 => DeviceClass::WirelessController.to_string(),
-                14 => DeviceClass::IntelligentController.to_string(),
-                15 => DeviceClass::SatelliteCommunicationsController.to_string(),
-                16 => DeviceClass::EncryptionController.to_string(),
-                17 => DeviceClass::SignalProcessingController.to_string(),
-                18 => DeviceClass::ProcessingAccelerator.to_string(),
-                19 => DeviceClass::NonEssentialInstrumentation.to_string(),
-                46 => DeviceClass::Coprocessor.to_string(),
-                255 => DeviceClass::Unassigned.to_string(),
-                _ => DeviceClass::Unclassified.to_string(),
-            }
+        self.class_name = match &self.class_id[0] {
+            1 => DeviceClass::MassStorageController.to_string(),
+            2 => DeviceClass::NetworkController.to_string(),
+            3 => DeviceClass::DisplayController.to_string(),
+            4 => DeviceClass::MultimediaController.to_string(),
+            5 => DeviceClass::MemoryController.to_string(),
+            6 => DeviceClass::Bridge.to_string(),
+            7 => DeviceClass::CommunicationController.to_string(),
+            8 => DeviceClass::GenericSystemPeripheral.to_string(),
+            9 => DeviceClass::InputDeviceController.to_string(),
+            10 => DeviceClass::DockingStation.to_string(),
+            11 => DeviceClass::Processor.to_string(),
+            12 => DeviceClass::SerialBusController.to_string(),
+            13 => DeviceClass::WirelessController.to_string(),
+            14 => DeviceClass::IntelligentController.to_string(),
+            15 => DeviceClass::SatelliteCommunicationsController.to_string(),
+            16 => DeviceClass::EncryptionController.to_string(),
+            17 => DeviceClass::SignalProcessingController.to_string(),
+            18 => DeviceClass::ProcessingAccelerator.to_string(),
+            19 => DeviceClass::NonEssentialInstrumentation.to_string(),
+            46 => DeviceClass::Coprocessor.to_string(),
+            255 => DeviceClass::Unassigned.to_string(),
+            _ => DeviceClass::Unclassified.to_string(),
         }
     }
 
     fn set_subclass_name(&mut self) {
-        // Look for line containing C & containing &self.class_id[1]
-        if let Ok(lines) = read_lines(PATH_TO_PCI_IDS) {
-            if !&self.class_id.is_empty() {
-                let class: [u8; 1] = [self.class_id[0]];
-                let encoded_class = hex::encode(&class);
-                let subclass: [u8; 1] = [self.class_id[1]];
-                let encoded_subclass = hex::encode(&subclass);
-                let mut found_my_class = false;
+        if self.class_id.is_empty() {
+            return;
+        }
 
-                for line in lines {
-                    if let Ok(l) = &line {
-                        if l.is_empty() || l.starts_with("#") {
-                            continue;
-                        } else if l.starts_with("C") && l.contains(&encoded_class) {
-                            found_my_class = true;
-                        } else if l.starts_with("\t")
-                            && l.contains(&encoded_subclass)
-                            && found_my_class
-                        {
-                            self.subclass_name = l.replace(&encoded_subclass, "").trim().to_owned();
-                            return;
-                        }
+        if let Ok(lines) = read_lines(PATH_TO_PCI_IDS) {
+            let class: [u8; 1] = [self.class_id[0]];
+            let encoded_class = hex::encode(&class);
+            let subclass: [u8; 1] = [self.class_id[1]];
+            let encoded_subclass = hex::encode(&subclass);
+            let mut found_my_class = false;
+
+            for line in lines {
+                if let Ok(l) = &line {
+                    if l.is_empty() || l.starts_with("#") {
+                        continue;
+                    } else if l.starts_with("C") && l.contains(&encoded_class) {
+                        found_my_class = true;
+                    } else if l.starts_with("\t") && l.contains(&encoded_subclass) && found_my_class
+                    {
+                        self.subclass_name = l.replace(&encoded_subclass, "").trim().to_owned();
+                        return;
                     }
                 }
             }
@@ -272,6 +306,10 @@ impl Properties for LinuxPCIDevice {
     }
 
     fn set_vendor_name(&mut self) {
+        if self.vendor_id.is_empty() {
+            return;
+        }
+
         if let Ok(lines) = read_lines(PATH_TO_PCI_IDS) {
             let ven = hex::encode(self.vendor_id.to_owned());
 
@@ -289,6 +327,10 @@ impl Properties for LinuxPCIDevice {
     }
 
     fn set_device_name(&mut self) {
+        if self.device_id.is_empty() {
+            return;
+        }
+
         if let Ok(lines) = read_lines(PATH_TO_PCI_IDS) {
             let dev = hex::encode(self.device_id.to_owned());
             for line in lines {
@@ -305,6 +347,10 @@ impl Properties for LinuxPCIDevice {
     }
 
     fn set_subsystem_name(&mut self) {
+        if self.subsystem_device_id.is_empty() {
+            return;
+        }
+
         if let Ok(lines) = read_lines(PATH_TO_PCI_IDS) {
             let sub_dev = hex::encode(self.subsystem_device_id.to_owned());
             let sub_ven = hex::encode(self.subsystem_vendor_id.to_owned());
@@ -369,6 +415,123 @@ impl Default for LinuxPCIDevice {
     }
 }
 
+impl Fetch for LinuxPCIDevice {
+    fn fetch(maximum_devices: Option<u8>) -> Vec<LinuxPCIDevice> {
+        let mut devices = Vec::new();
+        let entries = list_dir_entries(PATH_TO_PCI_DEVICES);
+        let mut i = 0u8;
+        for dir in entries {
+            if let Some(d) = dir.to_str() {
+                if let Some(m) = maximum_devices {
+                    i = i + 1;
+                    if i > m {
+                        continue;
+                    }
+                }
+
+                let device = LinuxPCIDevice::new(d);
+                devices.push(device);
+            }
+        }
+        return devices;
+    }
+
+    fn fetch_by_class(class: DeviceClass, maximum_devices: Option<u8>) -> Vec<LinuxPCIDevice> {
+        let mut devices = Vec::new();
+        let dir_entries = list_dir_entries(PATH_TO_PCI_DEVICES);
+        let mut i = 0u8;
+
+        for dir in dir_entries {
+            if let Some(d) = dir.to_str() {
+                if let Some(m) = maximum_devices {
+                    i = i + 1;
+                    if i > m {
+                        continue;
+                    }
+                }
+
+                // We're using `PCIDevice::reserved_new()` to initialize a PCIDevice
+                // with as little data as possible to avoid performance issues.
+                let mut device = LinuxPCIDevice::reserved_new(d);
+                if device.class_name() == class.to_string() {
+                    // We can now proceed to get and set the rest of the data
+                    // after having confirmed that the current PCIDevice's class matches
+                    // that provided by the user through a variant of the `DeviceClass` enum.
+                    device.set_address();
+                    device.set_vendor_id();
+                    device.set_device_id();
+                    device.set_numa_node();
+                    device.set_enabled();
+                    device.set_d3cold_allowed();
+                    device.set_revision();
+                    device.set_subsystem_device_id();
+                    device.set_subsystem_vendor_id();
+                    device.set_device_name();
+                    device.set_vendor_name();
+                    device.set_subsystem_name();
+                    device.set_subclass_name();
+                    devices.push(device);
+                }
+            }
+        }
+
+        return devices;
+    }
+
+    fn fetch_gpus(maximum_devices: Option<u8>) -> Vec<String> {
+        let mut gpus: Vec<String> = vec![];
+        let devices: Vec<LinuxPCIDevice> = Vec::new();
+        let dir_entries = list_dir_entries(PATH_TO_PCI_DEVICES);
+        let mut i = 0u8;
+
+        for dir in dir_entries {
+            if let Some(d) = dir.to_str() {
+                if let Some(m) = maximum_devices {
+                    i = i + 1;
+                    if i > m {
+                        continue;
+                    }
+                }
+
+                // We're using `PCIDevice::reserved_new()` to initialize a PCIDevice
+                // with as little data as possible to avoid performance issues.
+                let mut device = LinuxPCIDevice::reserved_new(d);
+                if device.class_name() == DeviceClass::DisplayController.to_string() {
+                    // We can now proceed to get and set the rest of the data
+                    // after having confirmed that the current PCIDevice's class matches
+                    // that provided by the user through a variant of the `DeviceClass` enum.
+                    device.set_enabled();
+                    // We're only going to return enabled gpus.
+                    if device.enabled {
+                        device.set_vendor_id();
+                        device.set_device_id();
+                        device.set_device_name();
+                        device.set_vendor_name();
+
+                        let whole_name = device.device_name();
+                        // Extracting text within brackets from device_name.
+                        if let Some(start_bytes) = whole_name.find("[") {
+                            if let Some(end_bytes) = whole_name.rfind("]") {
+                                device.device_name =
+                                    whole_name[start_bytes + 1..end_bytes].to_owned();
+                            }
+                        }
+
+                        if device.vendor_name().contains("Corporation") {
+                            device.vendor_name = device.vendor_name().replace(" Corporation", "");
+                        }
+
+                        let str = String::from(device.vendor_name + " " + &device.device_name);
+                        gpus.push(str);
+                    }
+                }
+            }
+        }
+
+        return gpus;
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -390,19 +553,19 @@ mod tests {
     #[test]
     fn test_class_id() {
         let device = LinuxPCIDevice::new(PLACEHOLDER_PCI_DEVICE);
-        assert_ne!(device.device_id(), "");
+        assert_ne!(device.device_id(), Vec::new());
     }
 
     #[test]
     fn test_vendor_id() {
         let device = LinuxPCIDevice::new(PLACEHOLDER_PCI_DEVICE);
-        assert_ne!(device.vendor_id(), "");
+        assert_ne!(device.vendor_id(), Vec::new());
     }
 
     #[test]
     fn test_device_id() {
         let device = LinuxPCIDevice::new(PLACEHOLDER_PCI_DEVICE);
-        assert_ne!(device.device_id(), "");
+        assert_ne!(device.device_id(), Vec::new());
     }
 
     #[test]
@@ -414,19 +577,19 @@ mod tests {
     #[test]
     fn test_revision() {
         let device = LinuxPCIDevice::new(PLACEHOLDER_PCI_DEVICE);
-        assert_ne!(device.revision(), "");
+        assert_ne!(device.revision(), Vec::new());
     }
 
     #[test]
     fn test_subsystem_vendor_id() {
         let device = LinuxPCIDevice::new(PLACEHOLDER_PCI_DEVICE);
-        assert_ne!(device.subsystem_vendor_id(), "");
+        assert_ne!(device.subsystem_vendor_id(), Vec::new());
     }
 
     #[test]
     fn test_subsystem_device_id() {
         let device = LinuxPCIDevice::new(PLACEHOLDER_PCI_DEVICE);
-        assert_ne!(device.subsystem_device_id(), "");
+        assert_ne!(device.subsystem_device_id(), Vec::new());
     }
 
     #[test]
